import json
import numpy as np
from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as plt
from sklearn.preprocessing import PolynomialFeatures
from scipy.linalg import lstsq
from datetime import datetime

from .helpers import shape_matcher, load_params, MyAMI_parameter_file
from .direct import calc_Fcorr

FCORR_COEFS = load_params('Fcorr_approx.json')

def approximate_Fcorr(TempC=25, Sal=35, Mg=0.0528171, Ca=0.0102821):
    """
    Approximate Fcorr using pre-calculated polynomial.
    
    Faster, but introduces up to ~0.2% error on correction factors in extreme

    Parameters
    ----------
    TempC : array-like
        Temperature in Celcius
    Sal : array-like
        Salinity in PSU
    Mg, Ca : array-like
        Mg and Ca concentration in mol/kg
    
    Returns
    -------
    dict
        Containing Fcorr factors for the specified inputs
    """

    print(
        'WARNING: using approximate MyAMI K correction factors\n' +
        'instead of calculated ones. These are only accurate to\n' +
        "within ~0.5%. Please don't use them for anything critical."
        )
    
    TempC, Sal, Mg, Ca = shape_matcher(TempC, Sal, Mg, Ca)

    check_limits(TempC, Sal, Mg, Ca)

    in_shape = TempC.shape
    
    # Build design matrix
    TempK = TempC + 273.15   
    X = np.vstack([TempK.ravel(), np.log(TempK.ravel()), Sal.ravel(), Mg.ravel(), Ca.ravel()]).T
    
    # This expands the input to a design matrix for 3rd-order polynomial with cross terms 
    # https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PolynomialFeatures.html
    poly = PolynomialFeatures(degree=3)
    X_ = poly.fit_transform(X)
    
    # calculate and return Fcorr
    return {k: X_.dot(c).reshape(in_shape) for k, c in FCORR_COEFS.items()}


def check_limits(TempC, Sal, Mg, Ca):

    warntext = (
        "\n\nDO NOT USE THESE APPROXIMATE CORRECTION FACTORS" +
        "\nPlease run a full MyAMI run instead."
    )

    if np.any((TempC < 0) | (TempC > 40)):
        raise ValueError('Temperature outside valid limits (0-40 C)' + warntext)

    if np.any((Sal < 30) | (Sal > 40)):
        raise ValueError('Salinity outside valid limits (30-40)' + warntext)

    if np.any((Mg < 0) | (Mg > 0.06)):
        raise ValueError('[Mg] outside valid limits (0-0.06)' + warntext)

    if np.any((Ca < 0) | (Ca > 0.06)):
        raise ValueError('[Ca] outside valid limits (0-0.06)' + warntext)


def generate_approximate_Fcorr_params(n=29, fit_reports=True):
    """
    Generate a Look Up Table (LUT) of parameters used to approximate Fcorr.

    The Fcorr factor is used to correct an empirical K value at
    a given TempC and Sal for Mg and Ca concentration, and should be
    applied as:

    Kcorr = Kempirical * Fcorr

    Parameter ranges are:
        TempC: 0 - 40 Celcius
        Sal: 30-40 PSU
        Mg, Ca: 0, 0.06 mol/kg

    Parameters
    ----------
    n : int
        The number of grid points to calculate for each parameter.
    
    Returns
    -------
    pandas.DataFrame
    """

    TempC = np.linspace(0, 40, n)
    Sal = np.linspace(30, 40, n)
    Mg = np.linspace(0, 0.06, n)
    Ca = np.linspace(0, 0.06, n)

    # grid inputs
    gTempC, gSal, gMg, gCa = np.meshgrid(TempC, Sal, Mg, Ca)
    
    # calculate Fcorr
    Fcorr = calc_Fcorr(Sal=gSal, TempC=gTempC, Mg=gMg, Ca=gCa)
    flat_Fcorr = {k: v.ravel() for k, v in Fcorr.items()}
    
    # generate design matrix for fitting
    X = np.vstack([
        gTempC.ravel() + 273.15,  # TempK
        np.log(gTempC.ravel() + 273.15),  # ln(TempK)
        gSal.ravel(),
        gMg.ravel(),
        gCa.ravel()
    ]).T
    
    poly = PolynomialFeatures(degree=3)
    X_ = poly.fit_transform(X)
    
    # calculate best-fit parameters
    coefs = {}
    for k in flat_Fcorr:
        y = flat_Fcorr[k]
        coefs[k], _, _, _ = lstsq(X_, y)
        
    if fit_reports:
        print(f'Fit reports saved in {MyAMI_parameter_file()}')
        pred = {k: X_.dot(c) for k, c in coefs.items()}
        for k in coefs:
            fname = MyAMI_parameter_file(f'Fcorr_approx_{k}.png')
            fig, axs = Fcorr_fit_report(k, flat_Fcorr[k], pred[k], X)
            fig.savefig(fname, dpi=150)

    with open(MyAMI_parameter_file('Fcorr_approx.json'), 'w') as f:
        json.dump({k: list(v) for k, v in coefs.items()}, f)


def Fcorr_fit_report(k, obs, pred, X):
    
    now = datetime.now().strftime('%Y/%m/%d %H:%M:%S')

    epc = 100 * (pred - obs) / obs  # percentage error

    kwargs = dict(s=2, alpha=0.2, color='k')

    axs = []

    gs = GridSpec(3, 2, wspace=0.1, hspace=0.3)
    fig = plt.figure(figsize=[7, 9])

    ax = fig.add_subplot(gs[0,:])
    ax.scatter(obs, epc, **kwargs)
    ax.set_xlabel('F_corr')
    ax.set_ylabel('% Error')
    ax.set_title(k, weight='bold', loc='left')
    ax.text(0.98, 0.95, f'Generated: {now}', transform=ax.transAxes, ha='right', va='top')
    axs.append(ax)

    ax = fig.add_subplot(gs[1,0])
    ax.scatter(X[:,0] - 273.15, epc, **kwargs)
    ax.set_xlabel('TempC')
    ax.set_ylabel('% Error')
    axs.append(ax)

    ax = fig.add_subplot(gs[1,1])
    ax.scatter(X[:,2], epc, **kwargs)
    ax.set_xlabel('Sal')
    ax.set_yticklabels([])
    axs.append(ax)

    ax = fig.add_subplot(gs[2,0])
    ax.scatter(X[:,3], epc, **kwargs)
    ax.set_xlabel('Mg')
    ax.set_ylabel('% Error')
    axs.append(ax)

    ax = fig.add_subplot(gs[2,1])
    ax.scatter(X[:,4], epc, **kwargs)
    ax.set_xlabel('Ca')
    ax.set_yticklabels([])
    axs.append(ax)

    return fig, axs


import numpy as np
from .helpers import expand_dims, match_dims, standard_seawater, calc_Istr, calc_KF, calc_KS

# TODO: new file for user-facing functions.

def calc_seawater_ions(Sal=35., Na=None, K=None, Ca=None, Mg=None, Sr=None, Cl=None, BOH4=None, HCO3=None, CO3=None, SO4=None):
    """
    Returns modern seawater composition with given ions modified at specified salinity. 

    All units are mol/kg.

    NOTE: Assumes that the provided ionic concentrations are at Sal=35.

    Returns
    -------
    tuple of arrays
        Containing (cations, anions) in the order:
        cations = [H, Na, K, Mg, Ca, Sr]
        anions = [OH, Cl, B(OH)4, HCO3, HSO4, CO3, SO4] 

    """
    modified_cations = [None, Na, K, Mg, Ca, Sr]
    modified_anions = [None, Cl, BOH4, HCO3, None, CO3, SO4]

    m_cations, m_anions = standard_seawater()

    m_cations = np.full(
        (m_cations.size, *Sal.shape),
        expand_dims(m_cations, Sal)
        )

    m_anions = np.full(
        (m_anions.size, *Sal.shape),
        expand_dims(m_anions, Sal)
        )

    for i, m in enumerate(modified_cations):
        if m is not None:
            m_cations[i] = m
    
    for i, m in enumerate(modified_anions):
        if m is not None:
            m_anions[i] = m

    sal_factor = Sal / 35.

    return m_cations * sal_factor, m_anions * sal_factor


def calculate_gKs(Tc, Sal, Na=None, K=None, Ca=None, Mg=None, Sr=None, Cl=None, BOH4=None, HCO3=None, CO3=None, SO4=None,
                  beta_0=None, beta_1=None, beta_2=None, C_phi=None, Theta_negative=None, Theta_positive=None, Phi_NNP=None, Phi_PPN=None, C1_HSO4=0):
    """
    Calculate Ks at given conditions using MyAMI model.
    """

    m_cation, m_anion = calc_seawater_ions(Sal, Na=Na, K=K, Mg=Mg, Ca=Ca, Sr=Sr, Cl=Cl, BOH4=BOH4, HCO3=HCO3, CO3=CO3, SO4=SO4)

    Istr = calc_Istr(Sal)

    [
        gamma_cation,
        gamma_anion,
        alpha_Hsws,
        alpha_Ht,
        alpha_OH,
        alpha_CO3,
    ] = CalculateGammaAndAlphas(
        Tc, Sal, Istr, m_cation, m_anion, 
        beta_0=beta_0, beta_1=beta_1, beta_2=beta_2, C_phi=C_phi,
        Theta_negative=Theta_negative, Theta_positive=Theta_positive,
        Phi_NNP=Phi_NNP, Phi_PPN=Phi_PPN, C1_HSO4=C1_HSO4)

    gammaT_OH = gamma_anion[0] * alpha_OH
    gammaT_BOH4 = gamma_anion[2]
    gammaT_HCO3 = gamma_anion[3]
    gammaT_CO3 = gamma_anion[5] * alpha_CO3

    # gammaT_Hsws = gamma_cation[0] * alpha_Hsws
    gammaT_Ht = gamma_cation[0] * alpha_Ht
    gammaT_Ca = gamma_cation[4]

    [gammaCO2, gammaCO2gas, gammaB] = gammaCO2_gammaB_fn(Tc, m_anion, m_cation)

    gKspC = 1 / gammaT_Ca / gammaT_CO3
    gKspA = 1 / gammaT_Ca / gammaT_CO3
    gK1 = 1 / gammaT_Ht / gammaT_HCO3 * gammaCO2
    gK2 = 1 / gammaT_Ht / gammaT_CO3 * gammaT_HCO3
    gKW = 1 / gammaT_Ht / gammaT_OH
    gKB = 1 / gammaT_BOH4 / gammaT_Ht * gammaB
    gK0 = 1 / gammaCO2 * gammaCO2gas
    gKHSO4 = 1 / gamma_anion[6] / gammaT_Ht * gamma_anion[4]

    return gKspC, gK1, gK2, gKW, gKB, gKspA, gK0, gKHSO4


def CalculateGammaAndAlphas(Tc, Sal, Istr, m_cation, m_anion,
                  beta_0=None, beta_1=None, beta_2=None, C_phi=None, Theta_negative=None, Theta_positive=None, Phi_NNP=None, Phi_PPN=None, C1_HSO4=0):
    """Calculate Gammas and Alphas for K calculations.

    Parameters
    ----------
    Tc : array-like
        Temperature in Celcius
    S : array-like
        Salinity in PSU
    Istr : array-like
        Ionic strength of solution
    m_cation : array-like
        Matrix of major cations in seawater in mol/kg in order:
        [H, Na, K, Mg, Ca, Sr]
    m_anion : array-like
        Matrix of major anions in seawater in mol/kg in order:
        [OH, Cl, B(OH)4, HCO3, HSO4, CO3, SO4]

    Returns
    -------
    list of arrays
        [gamma_cation, gamma_anion, alpha_Hsws, alpha_Ht, alpha_OH, alpha_CO3]
    """
    # Testbed case T=25C, I=0.7, seawatercomposition
    T = Tc + 273.15
    sqrtI = np.sqrt(Istr)
    
    # make tables of ion charges used in later calculations

    # cation order: [H, Na, K, Mg, Ca, Sr]
    cation_charges = np.array([1, 1, 1, 2, 2, 2])
    Z_cation = np.full(
        (cation_charges.size, *Tc.shape),
        expand_dims(cation_charges, Tc)
        )

    # anion order: [OH, Cl, B(OH)4, HCO3, HSO4, CO3, SO4]
    anion_charges = np.array([-1, -1, -1, -1, -1, -2, -2])
    Z_anion = np.full(
        (anion_charges.size, *Tc.shape),
        expand_dims(anion_charges, Tc)
        )   

    ##########################################################################

    A_phi = (
        3.36901532e-01
        - 6.32100430e-04 * T
        + 9.14252359 / T
        - 1.35143986e-02 * np.log(T)
        + 2.26089488e-03 / (T - 263)
        + 1.92118597e-6 * T * T
        + 4.52586464e01 / (680 - T)
    )  # note correction of last parameter, E + 1 instead of E-1
    # A_phi = 8.66836498e1 + 8.48795942e-2 * T - 8.88785150e-5 * T * T +
    # 4.88096393e-8 * T * T * T -1.32731477e3 / T - 1.76460172e1 * np.log(T)
    # # Spencer et al 1990

    f_gamma = -A_phi * (sqrtI / (1 + 1.2 * sqrtI) + (2 / 1.2) * np.log(1 + 1.2 * sqrtI))

    # E_cat = sum(m_cation * Z_cation)
    E_an = -sum(m_anion * Z_anion)
    E_cat = -E_an

    BMX_phi = beta_0 + beta_1 * np.exp(-2 * sqrtI)
    BMX = beta_0 + (beta_1 / (2 * Istr)) * (1 - (1 + 2 * sqrtI) * np.exp(-2 * sqrtI))
    BMX_apostroph = (beta_1 / (2 * Istr * Istr)) * (-1 + (1 + (2 * sqrtI) + (2 * sqrtI)) * np.exp(-2 * sqrtI))
    CMX = C_phi / (2 * np.sqrt(-np.expand_dims(Z_anion, 0) * np.expand_dims(Z_cation, 1)))
    
    # BMX* and CMX are calculated differently for 2:2 ion pairs, corrections
    # below  # § alpha2= 6 for borates ... see Simonson et al 1988
    
    # MgBOH42
    cat, an = 3, 2
    BMX_phi[cat, an] = (
        beta_0[cat, an]
        + beta_1[cat, an] * np.exp(-1.4 * sqrtI)
        + beta_2[cat, an] * np.exp(-6 * sqrtI)
    )
    BMX[cat, an] = (
        beta_0[cat, an]
        + (beta_1[cat, an] / (0.98 * Istr))
        * (1 - (1 + 1.4 * sqrtI) * np.exp(-1.4 * sqrtI))
        + (beta_2[cat, an] / (18 * Istr)) * (1 - (1 + 6 * sqrtI) * np.exp(-6 * sqrtI))
    )
    BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr * Istr)) * (
        -1 + (1 + 1.4 * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)
    ) + (beta_2[cat, an] / (18 * Istr)) * (
        -1 - (1 + 6 * sqrtI + 18 * Istr) * np.exp(-6 * sqrtI)
    )
    
    # MgSO4
    cat, an = 3, 6 
    BMX_phi[cat, an] = (
        beta_0[cat, an]
        + beta_1[cat, an] * np.exp(-1.4 * sqrtI)
        + beta_2[cat, an] * np.exp(-12 * sqrtI)
    )
    BMX[cat, an] = (
        beta_0[cat, an]
        + (beta_1[cat, an] / (0.98 * Istr))
        * (1 - (1 + 1.4 * sqrtI) * np.exp(-1.4 * sqrtI))
        + (beta_2[cat, an] / (72 * Istr)) * (1 - (1 + 12 * sqrtI) * np.exp(-12 * sqrtI))
    )
    BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr * Istr)) * (
        -1 + (1 + 1.4 * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)
    ) + (beta_2[cat, an] / (72 * Istr * Istr)) * (
        -1 - (1 + 12 * sqrtI + 72 * Istr) * np.exp(-12 * sqrtI)
    )
    # BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr)) * (-1 + (1 + 1.4
    # * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)) + (beta_2[cat, an] / (72 *
    # Istr)) * (-1-(1 + 12 * sqrtI + 72 * Istr) * np.exp(-12 * sqrtI)) # not 1 /
    # (0.98 * Istr * Istr) ... compare M&P98 equation A17 with Pabalan and Pitzer
    # 1987 equation 15c / 16b
    
    # CaBOH42
    cat, an = 4, 2 
    BMX_phi[cat, an] = (
        beta_0[cat, an]
        + beta_1[cat, an] * np.exp(-1.4 * sqrtI)
        + beta_2[cat, an] * np.exp(-6 * sqrtI)
    )
    BMX[cat, an] = (
        beta_0[cat, an]
        + (beta_1[cat, an] / (0.98 * Istr))
        * (1 - (1 + 1.4 * sqrtI) * np.exp(-1.4 * sqrtI))
        + (beta_2[cat, an] / (18 * Istr)) * (1 - (1 + 6 * sqrtI) * np.exp(-6 * sqrtI))
    )
    BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr * Istr)) * (
        -1 + (1 + 1.4 * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)
    ) + (beta_2[cat, an] / (18 * Istr)) * (
        -1 - (1 + 6 * sqrtI + 18 * Istr) * np.exp(-6 * sqrtI)
    )
    
    # CaSO4
    cat, an = 4, 6
    BMX_phi[cat, an] = (
        beta_0[cat, an]
        + beta_1[cat, an] * np.exp(-1.4 * sqrtI)
        + beta_2[cat, an] * np.exp(-12 * sqrtI)
    )
    BMX[cat, an] = (
        beta_0[cat, an]
        + (beta_1[cat, an] / (0.98 * Istr))
        * (1 - (1 + 1.4 * sqrtI) * np.exp(-1.4 * sqrtI))
        + (beta_2[cat, an] / (72 * Istr)) * (1 - (1 + 12 * sqrtI) * np.exp(-12 * sqrtI))
    )
    BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr * Istr)) * (
        -1 + (1 + 1.4 * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)
    ) + (beta_2[cat, an] / (72 * Istr)) * (
        -1 - (1 + 12 * sqrtI + 72 * Istr) * np.exp(-12 * sqrtI)
    )

    # SrBOH42
    cat, an = 5, 2 
    BMX_phi[cat, an] = (
        beta_0[cat, an]
        + beta_1[cat, an] * np.exp(-1.4 * sqrtI)
        + beta_2[cat, an] * np.exp(-6 * sqrtI)
    )
    BMX[cat, an] = (
        beta_0[cat, an]
        + (beta_1[cat, an] / (0.98 * Istr))
        * (1 - (1 + 1.4 * sqrtI) * np.exp(-1.4 * sqrtI))
        + (beta_2[cat, an] / (18 * Istr)) * (1 - (1 + 6 * sqrtI) * np.exp(-6 * sqrtI))
    )
    BMX_apostroph[cat, an] = (beta_1[cat, an] / (0.98 * Istr * Istr)) * (
        -1 + (1 + 1.4 * sqrtI + 0.98 * Istr) * np.exp(-1.4 * sqrtI)
    ) + (beta_2[cat, an] / (18 * Istr)) * (
        -1 - (1 + 6 * sqrtI + 18 * Istr) * np.exp(-6 * sqrtI)
    )

    # H-SO4
    cat, an = 0, 6
    # BMX* is calculated with T-dependent alpha for H-SO4; see Clegg et al.,
    # 1994 --- Millero and Pierrot are completly off for this ion pair
    xClegg = (2 - 1842.843 * (1 / T - 1 / 298.15)) * sqrtI
    # xClegg = (2) * sqrtI
    gClegg = 2 * (1 - (1 + xClegg) * np.exp(-xClegg)) / (xClegg * xClegg)
    # alpha = (2 - 1842.843 * (1 / T - 1 / 298.15)) see Table 6 in Clegg et al
    # 1994
    BMX[cat, an] = beta_0[cat, an] + beta_1[cat, an] * gClegg
    BMX_apostroph[cat, an] = beta_1[cat, an] / Istr * (np.exp(-xClegg) - gClegg)

    CMX[cat, an] = C_phi[cat, an] + 4 * C1_HSO4 * (
        6
        - (6 + 2.5 * sqrtI * (6 + 3 * 2.5 * sqrtI + 2.5 * sqrtI * 2.5 * sqrtI))
        * np.exp(-2.5 * sqrtI)
    ) / (
        2.5 * sqrtI * 2.5 * sqrtI * 2.5 * sqrtI * 2.5 * sqrtI
    )  # w = 2.5 ... see Clegg et al., 1994

    # unusual alpha=1.7 for Na2SO4
    # BMX[1, 6] = beta_0[1, 6] + (beta_1[1, 6] / (2.89 * Istr)) * 2 * (1 - (1 + 1.7 * sqrtI) * np.exp(-1.7 * sqrtI))
    # BMX[1, 6] = beta_0[1, 6] + (beta_1[1, 6] / (1.7 * Istr)) * (1 - (1 + 1.7 * sqrtI) * np.exp(-1.7 * sqrtI))

    # BMX[4, 6] =BMX[4, 6] * 0  # knock out Ca-SO4
    
    R = (m_anion * np.expand_dims(m_cation, 1) * BMX_apostroph).sum((0,1))
    S = (m_anion * np.expand_dims(m_cation, 1) * CMX).sum((0,1))

    # ln_gammaCl = Z_anion[1] * Z_anion[1] * f_gamma + R - S

    # Original ln_gamma_anion calculation loop:
    ln_gamma_anion = Z_anion * Z_anion * (f_gamma + R) + Z_anion * S
    for an in range(0, 7):
        for cat in range(0, 6):
            ln_gamma_anion[an] += 2 * m_cation[cat] * (
                BMX[cat, an] + E_cat * CMX[cat, an]
            )
        for an2 in range(0, 7):
            ln_gamma_anion[an] += m_anion[an2] * (
                2 * Theta_negative[an, an2]
            )
        for an2 in range(0, 7):
            for cat in range(0, 6):
                ln_gamma_anion[an] += (
                    m_anion[an2] * m_cation[cat] * Phi_NNP[an, an2, cat]
                )
        for cat in range(0, 6):
            for cat2 in range(cat + 1, 6):
                ln_gamma_anion[an] += (
                    m_cation[cat] * m_cation[cat2] * Phi_PPN[cat, cat2, an]
                )
    
    # vectorised ln_gamma_anion calculation:
    # cat, cat2 = np.triu_indices(6, 1)
    # ln_gamma_anion = (
    #     Z_anion * Z_anion * (f_gamma + R) + Z_anion * S + 
    #     (2 * np.expand_dims(m_cation, 1) * (BMX + E_cat * CMX)).sum(0) + 
    #     (np.expand_dims(m_anion, 1) * 2 * Theta_negative).sum(0) + 
    #     (np.expand_dims(m_anion, (0,2)) * np.expand_dims(m_cation, (0,1)) * Phi_NNP).sum(axis=(1,2)) +
    #     (np.expand_dims(m_cation[cat], 1) * np.expand_dims(m_cation[cat2], 1) * Phi_PPN[cat, cat2]).sum(axis=0)
    # )  # TODO - could be simplified further?
    gamma_anion = np.exp(ln_gamma_anion)


    # ln_gammaCl = Z_anion[1] * Z_anion[1] * f_gamma + R - S

    # Original ln_gamma_cation calculation loop:
    ln_gamma_cation = Z_cation * Z_cation * (f_gamma + R) + Z_cation * S
    for cat in range(0, 6):
        for an in range(0, 7):
            ln_gamma_cation[cat] += 2 * m_anion[an] * (
                BMX[cat, an] + E_cat * CMX[cat, an]
            )
        for cat2 in range(0, 6):
            ln_gamma_cation[cat] += m_cation[cat2] * (2 * Theta_positive[cat, cat2])
        for cat2 in range(0, 6):
            for an in range(0, 7):
                ln_gamma_cation[cat] += (
                    m_cation[cat2] * m_anion[an] * Phi_PPN[cat, cat2, an]
                )
        for an in range(0, 7):
            for an2 in range(an + 1, 7):
                ln_gamma_cation[cat] += (
                    + m_anion[an] * m_anion[an2] * Phi_NNP[an, an2, cat]
                )

    # vectorised ln_gamma_cation calculation:
    # an, an2 = np.triu_indices(7, 1)
    # ln_gamma_cation = (
    #     Z_cation * Z_cation * (f_gamma + R) + Z_cation * S +
    #     (2 * np.expand_dims(m_anion, 0) * (BMX + E_cat * CMX)).sum(axis=1) +
    #     (np.expand_dims(m_cation, 1) * (2 * Theta_positive)).sum(axis=0) +
    #     (np.expand_dims(m_cation, (0,2)) * np.expand_dims(m_anion, (0,1)) * Phi_PPN).sum(axis=(1,2))+
    #     (np.expand_dims(m_anion[an], 1) * np.expand_dims(m_anion[an2], 1) * Phi_NNP[an, an2]).sum(axis=0)
    # )  # TODO - could be simplified further?
    gamma_cation = np.exp(ln_gamma_cation)

    # choice of pH-scale = total pH-scale [H]T = [H]F + [HSO4]
    # so far gamma_H is the [H]F activity coefficient (= free-H pH-scale)
    # thus, conversion is required
    # * (gamma_anion[4] / gamma_anion[6] / gamma_cation[0])
    
    K_HSO4_conditional = calc_KS(TK=T, Sal=Sal, Istr=Istr)
    K_HF_conditional = calc_KF(TK=T, Sal=Sal)
    
    # print (gamma_anion[4], gamma_anion[6], gamma_cation[0])
    # alpha_H = 1 / (1+ m_anion[6] / K_HSO4_conditional + 0.0000683 / (7.7896E-4 * 1.1 / 0.3 / gamma_cation[0]))
    alpha_Hsws = 1 / (
        # 1 + m_anion[6] / K_HSO4_conditional + 0.0000683 / (supplyKHF(T, sqrtI))
        1 + m_anion[6] / K_HSO4_conditional + 0.0000683 / K_HF_conditional
    )
    alpha_Ht = 1 / (1 + m_anion[6] / K_HSO4_conditional)
    # alpha_H = 1 / (1+ m_anion[6] / K_HSO4_conditional)

    # A number of ion pairs are calculated explicitly: MgOH, CaCO3, MgCO3, SrCO3
    # since OH and CO3 are rare compared to the anions the anion alpha (free /
    # total) are assumed to be unity
    gamma_MgCO3 = 1
    gamma_CaCO3 = gamma_MgCO3
    gamma_SrCO3 = gamma_MgCO3

    b0b1CPhi_MgOH = np.array([-0.1, 1.658, 0, 0.028])
    BMX_MgOH = b0b1CPhi_MgOH[0] + (b0b1CPhi_MgOH[1] / (2 * Istr)) * (
        1 - (1 + 2 * sqrtI) * np.exp(-2 * sqrtI)
    )
    ln_gamma_MgOH = 1 * (f_gamma + R) + (1) * S
    # interaction between MgOH-Cl affects MgOH gamma
    ln_gamma_MgOH = ln_gamma_MgOH + 2 * m_anion[1] * (
        BMX_MgOH + E_cat * b0b1CPhi_MgOH[2]
    )
    # interaction between MgOH-Mg-OH affects MgOH gamma
    ln_gamma_MgOH = ln_gamma_MgOH + m_cation[3] * m_anion[1] * b0b1CPhi_MgOH[3]
    gamma_MgOH = np.exp(ln_gamma_MgOH)

    K_MgOH = np.power(10, -(3.87 - 501.6 / T)) / (
        gamma_cation[3] * gamma_anion[0] / gamma_MgOH
    )
    K_MgCO3 = np.power(10, -(1.028 + 0.0066154 * T)) / (
        gamma_cation[3] * gamma_anion[5] / gamma_MgCO3
    )
    K_CaCO3 = np.power(10, -(1.178 + 0.0066154 * T)) / (
        gamma_cation[4] * gamma_anion[5] / gamma_CaCO3
    )
    # K_CaCO3 = np.power(10, (-1228.732 - 0.299444 * T + 35512.75 / T +485.818 * np.log10(T))) / (gamma_cation[4] * gamma_anion[5] / gamma_CaCO3) # Plummer and Busenberg82
    # K_MgCO3 = np.power(10, (-1228.732 +(0.15) - 0.299444 * T + 35512.75 / T
    # +485.818 * np.log10(T))) / (gamma_cation[4] * gamma_anion[5] /
    # gamma_CaCO3)# Plummer and Busenberg82
    K_SrCO3 = np.power(10, -(1.028 + 0.0066154 * T)) / (
        gamma_cation[5] * gamma_anion[5] / gamma_SrCO3
    )

    alpha_OH = 1 / (1 + (m_cation[3] / K_MgOH))
    alpha_CO3 = 1 / (
        1 + (m_cation[3] / K_MgCO3) + (m_cation[4] / K_CaCO3) + (m_cation[5] / K_SrCO3)
    )

    return gamma_cation, gamma_anion, alpha_Hsws, alpha_Ht, alpha_OH, alpha_CO3

def gammaCO2_gammaB_fn(Tc, m_an, m_cat):
    T = Tc + 273.15
    lnT = np.log(T)

    m_ion = np.array(
        [m_cat[0], m_cat[1], m_cat[2], m_cat[3], m_cat[4], m_an[1], m_an[6]]
    )

    param_lamdaCO2 = np.zeros([7, 5])
    param_lamdaCO2[0, :] = [0, 0, 0, 0, 0]  # H
    param_lamdaCO2[1, :] = [
        -5496.38465,
        -3.326566,
        0.0017532,
        109399.341,
        1047.021567,
    ]  # Na
    param_lamdaCO2[2, :] = [
        2856.528099,
        1.7670079,
        -0.0009487,
        -55954.1929,
        -546.074467,
    ]  # K
    param_lamdaCO2[3, :] = [
        -479.362533,
        -0.541843,
        0.00038812,
        3589.474052,
        104.3452732,
    ]  # Mg
    # param_lamdaCO2[3, :] = [9.03662673e+03, 5.08294701e+00, -2.51623005e-03, -1.88589243e+05, -1.70171838e+03]  # Mg refitted
    param_lamdaCO2[4, :] = [
        -12774.6472,
        -8.101555,
        0.00442472,
        245541.5435,
        2452.50972,
    ]  # Ca
    # param_lamdaCO2[4, :] = [-8.78153999e+03, -5.67606538e+00, 3.14744317e-03, 1.66634223e+05, 1.69112982e+03]  # Ca refitted
    param_lamdaCO2[5, :] = [
        1659.944942,
        0.9964326,
        -0.00052122,
        -33159.6177,
        -315.827883,
    ]  # Cl
    param_lamdaCO2[6, :] = [
        2274.656591,
        1.8270948,
        -0.00114272,
        -33927.7625,
        -457.015738,
    ]  # SO4

    param_zetaCO2 = np.zeros([2, 6, 5])
    param_zetaCO2[0, 0, :] = [
        -804.121738,
        -0.470474,
        0.000240526,
        16334.38917,
        152.3838752,
    ]  # Cl & H
    param_zetaCO2[0, 1, :] = [
        -379.459185,
        -0.258005,
        0.000147823,
        6879.030871,
        73.74511574,
    ]  # Cl & Na
    param_zetaCO2[0, 2, :] = [
        -379.686097,
        -0.257891,
        0.000147333,
        6853.264129,
        73.79977116,
    ]  # Cl & K
    param_zetaCO2[0, 3, :] = [
        -1342.60256,
        -0.772286,
        0.000391603,
        27726.80974,
        253.62319406,
    ]  # Cl & Mg
    param_zetaCO2[0, 4, :] = [
        -166.06529,
        -0.018002,
        -0.0000247349,
        5256.844332,
        27.377452415,
    ]  # Cl & Ca
    param_zetaCO2[1, 1, :] = [
        67030.02482,
        37.930519,
        -0.0189473,
        -1399082.37,
        -12630.27457,
    ]  # SO4 & Na
    param_zetaCO2[1, 2, :] = [
        -2907.03326,
        -2.860763,
        0.001951086,
        30756.86749,
        611.37560512,
    ]  # SO4 & K
    param_zetaCO2[1, 3, :] = [
        -7374.24392,
        -4.608331,
        0.002489207,
        143162.6076,
        1412.302898,
    ]  # SO4 & Mg

    lamdaCO2 = np.zeros((7, *Tc.shape))
    for ion in range(0, 7):
        lamdaCO2[ion] = (
            param_lamdaCO2[ion, 0]
            + param_lamdaCO2[ion, 1] * T
            + param_lamdaCO2[ion, 2] * T ** 2
            + param_lamdaCO2[ion, 3] / T
            + param_lamdaCO2[ion, 4] * lnT
        )

    zetaCO2 = np.zeros([2, 5, *Tc.shape])
    for ion in range(0, 5):
        zetaCO2[0, ion] = (
            param_zetaCO2[0, ion, 0]
            + param_zetaCO2[0, ion, 1] * T
            + param_zetaCO2[0, ion, 2] * T ** 2
            + param_zetaCO2[0, ion, 3] / T
            + param_zetaCO2[0, ion, 4] * lnT
        )
    for ion in range(1, 4):
        zetaCO2[1, ion] = (
            param_zetaCO2[1, ion, 0]
            + param_zetaCO2[1, ion, 1] * T
            + param_zetaCO2[1, ion, 2] * T ** 2
            + param_zetaCO2[1, ion, 3] / T
            + param_zetaCO2[1, ion, 4] * lnT
        )

    
    # original calculation:
    # ln_gammaCO2 = 0
    # for ion in range(0, 7):
    #     ln_gammaCO2 = ln_gammaCO2 + m_ion[ion] * 2 * lamdaCO2[ion]
    
    # vectorised calculation:
    ln_gammaCO2 = (m_ion * 2 * lamdaCO2).sum(0)

    # for cat in range(0, 5):
    # ln_gammaCO2 = ln_gammaCO2 + m_ion[5] * m_ion[cat] * zetaCO2[0, cat] + m_ion[6] * m_ion[cat] * zetaCO2[1, cat]

    gammaCO2 = np.exp(ln_gammaCO2)  # as according to He and Morse 1993
    # gammaCO2 = np.power(10, ln_gammaCO2) # pK1 is "correct if log-base 10 is assumed

    gammaCO2gas = np.exp(
        1
        / (
            8.314462175
            * T
            * (0.10476 - 61.0102 / T - 660000 / T / T / T - 2.47e27 / np.power(T, 12))
        )
    )

    ##########################
    # CALCULATION OF gammaB
    lamdaB = np.array([0, -0.097, -0.14, 0, 0, 0.091, 0.018])  # Felmy and Wear 1986
    # lamdaB = np.array([0.109, 0.028, -0.026, 0.191, 0.165, 0, -0.205]) #Chanson and Millero 2006
    
    # original calculation:

    # ln_gammaB = m_ion[1] * m_ion[6] * 0.046  # tripple ion interaction Na-SO4
    # for ion in range(0, 7):
    #     ln_gammaB = ln_gammaB + m_ion[ion] * 2 * lamdaB[ion]
    
    # vectorised calculation:
    ln_gammaB = m_ion[1] * m_ion[6] * 0.046 + (m_ion * 2 * match_dims(lamdaB, m_ion)).sum(0)
    
    gammaB = np.exp(ln_gammaB)  # as according to Felmy and Wear 1986
    # print gammaB

    return gammaCO2, gammaCO2gas, gammaB


"""
Helper functions used throughout MyAMI.
"""
import json
import numpy as np
import pkg_resources as pkgrs

def MyAMI_parameter_file(fname=''):
    return pkgrs.resource_filename('pymyami', f'parameters/{fname}')

def expand_dims(orig, target):
    """
    Adds additional dimensions to orig so it can be broadcast on target.
    """
    return np.expand_dims(orig, tuple(range(orig.ndim, target.ndim + orig.ndim)))

def match_dims(orig, target):
    """
    Adds additional dimensions to orig to match the number of dimensions in target.
    """
    return np.expand_dims(orig, tuple(range(orig.ndim, target.ndim)))

def shape_matcher(*args):
    """
    Returns all given arrays as the same shape.
    """
    biggest = 0
    shape = None
    for a in args:
        if a is not None:
            a = np.asanyarray(a)
            if a.size > biggest:
                biggest = a.size
                shape = a.shape
    
    for a in args:
        if a is not None:
            a = np.asanyarray(a)
            if a.shape != shape:
                yield np.full(shape, a)
            else:
                yield a
        else:
            yield a

def load_params(param_file, asarrays=True):
    """
    Load parameters from a json file and convert the entries to np.ndarray.

    Parameters
    ----------
    param_file : string
        The name of the file within MyAMI/resources to load.
    asarrays : bool, optional
        If true, entries are converted to np.ndarray, by default True

    Returns
    -------
    dict
        The contents of the json file as a dict.
    """
    with open(MyAMI_parameter_file(param_file), 'r') as f:
        params = json.load(f)
    
    if asarrays:
        return {k: np.array(v) for k, v in params.items()}
    return params

def calc_Istr(Sal):
    """
    Calculation ionic strength from Salinity
    """
    return 19.924 * Sal / (1000 - 1.005 * Sal)

def standard_seawater(S=35.):
    """
    Return modern seawater ionic composition at specified salinity
    in units of mol/kg.

    Parameters
    ----------
    S : array-like
        Salinity in PSU

    Returns
    -------
    tuple of arrays
        Containing (cations, anions) in the order:
        cations = [H, Na, K, Mg, Ca, Sr]
        anions = [OH, Cl, B(OH)4, HCO3, HSO4, CO3, SO4] 
    """

    cation_concs = np.array([
        0.00000001,  # H ion; pH of about 8
        0.4689674,  # Na Millero et al., 2008; Dickson OA-guide
        0.0102077,  # K Millero et al., 2008; Dickson OA-guide
        0.0528171,  # Mg Millero et al., 2008; Dickson OA-guide
        0.0102821,  # Ca Millero et al., 2008; Dickson OA-guide
        0.0000907  # Sr Millero et al., 2008; Dickson OA-guide
    ]) * S / 35.

    anion_concs = np.array([
        0.0000010,  # OH ion; pH of about 8
        0.5458696,  # Cl Millero et al., 2008; Dickson OA-guide
        0.0001008,  # BOH4 Millero et al., 2008; Dickson OA-guide; pH of about 8 -- borate,
        0.0017177,  # HCO3 Millero et al., 2008; Dickson OA-guide
        0.0282352 * 1e-6,  # HSO4 Millero et al., 2008; Dickson OA-guide
        0.0002390,  # CO3 Millero et al., 2008; Dickson OA-guide
        0.0282352  # SO4 Millero et al., 2008; Dickson OA-guide
    ]) * S / 35.

    return cation_concs, anion_concs

def calc_KS(TK=25., Sal=35., Istr=None):
    lnTK = np.log(TK)
    p = [141.328, -4276.1, -23.093, -13856, 324.57, -47.986, 35474, -771.54, 114.723, -2698, 1776]
    
    return np.exp(
        p[0]
        + p[1] / TK
        + p[2] * lnTK
        + np.sqrt(Istr) * (p[3] / TK + p[4] + p[5] * lnTK)
        + Istr * (p[6] / TK + p[7] + p[8] * lnTK)
        + p[9] / TK * Istr * np.sqrt(Istr)
        + p[10] / TK * Istr ** 2
        + np.log(1 - 0.001005 * Sal)
    )
    
def calc_KF(TK=25., Sal=35.):
    p = [874, -9.68, 0.111]  # from Best Practices guide.
    # p = [1590.2, 12.641, 1.525]  # used in MyAMI_V1
    
    return np.exp(
        p[0] / TK + 
        p[1] + 
        p[2] * Sal**0.5
    )


import os
import re
import numpy as np
import pandas as pd
from glob import glob
from .helpers import MyAMI_parameter_file, expand_dims, match_dims, load_params

# dictionaries of ions containing their matrix indices
# positive ions H+=0; Na+=1; K+=2; Mg2+=3; Ca2+=4; Sr2+=5
CA_IND = {
    'H': 0,
    'Na': 1,
    'K': 2,
    'Mg': 3,
    'Ca': 4,
    'Sr': 5
}

# negative ions  OH-=0; Cl-=1; B(OH)4-=2; HCO3-=3; HSO4-=4; CO3-=5; SO4-=6;
AN_IND = {
    'OH': 0,
    'Cl': 1,
    'B(OH)4': 2,
    'HCO3': 3,
    'HSO4': 4,
    'CO3': 5,
    'SO4': 6
}

# count ions
N_CA = len(CA_IND)  # H+=0; Na+=1; K+=2; Mg2+=3; Ca2+=4; Sr2+=5
N_AN = len(AN_IND)  # OH-=0; Cl-=1; B(OH)4-=2; HCO3-=3; HSO4-=4; CO3-=5; SO4-=6;

# build a regex for pulling ions out of salt names
recations = '|'.join(CA_IND.keys())
reanions = '|'.join(AN_IND.keys())
reanions = reanions.replace('(', '\(').replace(')', '\)')  # escape brackets
sm = re.compile('^(' + recations + ')[0-9]?\(?(' + reanions + ')\)?[0-9]?$')

def break_salt(s):
    """
    Returns positive and negative species names from salt or ion pair (X-Y).
    """
    if '-' in s:
        return s.split('-')
    
    match = sm.match(s)
    if match:
        gs = match.groups()
        return gs[0], gs[1]
    else:
        return None, None

# dictionary containing all valid ions
Iind = CA_IND.copy()
Iind.update(AN_IND)

# helper functions for converting tables into calculationg matrices

def filter_terms(tab, valid_ions):
    include = []
    for ions in tab.Pair.str.split('-'):
        include.append(~np.any([i not in valid_ions for i in ions]))

    return tab.loc[include]

def get_ion_index(ions):
    return tuple([Iind[k] for k in ions.split('-')])

# Load Parameter Tables for calculating pitzer parameters
TABLES = {}
fs = glob(MyAMI_parameter_file('Tab*.csv'))
for f in fs:
    fname = os.path.split(f)[-1].replace('.csv', '')
    TABLES[fname] = pd.read_csv(f, comment='#')
    TABLES[fname].fillna(0, inplace=True)

# remove unused pairs
TABA11 = filter_terms(TABLES['TabA11'], Iind)
TABA10 = filter_terms(TABLES['TabA10'], Iind)


# Table A1-9 equations for beta and C calculation
def EqA1(a, TK, TKinv, lnTK, **kwargs):
    # T in Kelvin
    return (
        a[0] +
        a[1] * TK +
        a[2] * TKinv +
        a[3] * lnTK +
        a[4] / (TK - 263) +
        a[5] * TK**2 +
        a[6] / (680 - TK) +
        a[7] / (TK - 227)
    )

def EqA2(a, TK, **kwargs):
    return a[0] + a[1] * TK + a[2] * TK**2

def EqA3A4(a, Tsub, **kwargs):
    return a[0] + a[1] * Tsub + a[2] * Tsub**2

def EqA5A6(a, TK, Tsub, **kwargs):
    return a[0] + a[1] * Tsub + a[2] * (TK - 303.15)**2

def EqA7(a, TK, **kwargs):
    PR, PJ, PLR = a[0:3]
    a = 88893.4225
    b = 8834524.63945833
    return (
        PR + PJ * (b - a * PLR) * (1 / TK - (1 / 298.15))
        + PJ / 6 * (TK**2 - a)
    )

# link tables to equations
EQ_TABLES = {
    'TabA1': EqA1,
    'TabA2': EqA2,
    'TabA3': EqA3A4,
    'TabA4': EqA3A4,
    'TabA5': EqA5A6,
    'TabA6': EqA5A6,
    'TabA7': EqA7,
    'TabA9': None,  # both special
    'TabSpecial': None  # special cases handled below
}

# Special Case Equations noted in table subscripts
def EqA2_MgSO4(a, TK, **kwargs):
    return (
        a[0] * ((TK / 2) + (88804) / (2 * TK) - 298)
        + a[1] * ((TK**2 / 6) + (26463592) / (3 * TK) - (88804 / 2))
        + a[2] * (TK**3 / 12 + 88804 * 88804 / (4 * TK) - 26463592 / 3)
        + a[3]
        * ((TK**4 / 20) + 88804 * 26463592 / (5 * TK) - 88804 * 88804 / 4)
        + a[4] * (298 - (88804 / TK))
        + a[5]
    )

def EqA3A4_XHS(a, Tsub, TKinv, **kwargs):
    return a[0] + a[1] * TKinv + a[2] * Tsub**2

def EqA3_XSO3(a, TK, TKinv):
    return a[0] + a[1] * (TKinv - 1/298.15) + a[2] * np.log(TK / 298.15)

def EqA9_HCl(a, TK, **kwargs):
    return a[0] + a[1] * TK + a[2] / TK

def EqA9_HSO4(a, TK, **kwargs):
    TKsub3 = TK - 328.15
    return (
        a[0] + TKsub3 * 1e-3 * 
        (a[1] + TKsub3 * ((a[2] / 2) + TKsub3 * (a[3] / 6)))
        )

# dictionary of special cases
EQ_SPECIAL = {
    'TabA2': {
        'MgSO4': EqA2_MgSO4
    },
    'TabA3': {
        'KHS': EqA3A4_XHS,
        'Na2SO3': EqA3_XSO3,
        'NaHSO3': EqA3_XSO3,
    },
    'TabA4': {
        'KHS': EqA3A4_XHS,
    },
    'TabA9': {
        'H-Cl': EqA9_HCl,
        'H-SO4': EqA9_HSO4
    },
}

# Special case equations used in Mathis Hain's code.
def EQ_SPECIAL_Na2SO4_Moller(a, TK, lnTK, **kwargs):
    return (
        a[0] +
        a[1] * TK +
        a[2] / TK +
        a[3] * lnTK +
        a[4] / (TK - 263) +
        a[5] * TK**2 +
        a[6] / (680. - TK)
    )

def EQ_SPECIAL_MgHSO4(a, Tsub, **kwargs):
    return a[0] + a[1] * Tsub

# add these to dictionary
EQ_SPECIAL['TabSpecial'] = {
        'Na2SO4': EQ_SPECIAL_Na2SO4_Moller,
        'Mg(HSO4)2': EQ_SPECIAL_MgHSO4
    }

# Phi and Theta Equation
def EqA10(a, TK):
    """
    Calculate Phi and Theta parameters as a function of TK accoring to 
    """
    # a1 + a2 / T + a3 * T + a4 * (T - 298.15) + a5 * (T - 298.15)**2
    TKsub = TK - 298.15
    return a[0] + a[1] / TK + a[2] * 1e-4 * TK + a[3] * 1e-4 * TKsub + a[4] * 1e-6 * TKsub**2

###############################
# Matrix Construction Functions
###############################

def calc_beta_C(TK):
    """
    Calculate matrices of beta_0, beta_1, beta_2 and C_phi at given TK.

    Matrices are constructed from tables A1-A9 of Millero and Pierrot (1998; 
    doi:10.1023/A:1009656023546), with modifications implemented by Hain et 
    al (2015).

    Parameters
    ----------
    TK : array-like
        Temperature in Kelvin
    
    Returns
    -------
    dict
        Containing {beta_0, beta_1, beta_2, C_phi}
    """
    TKinv = 1. / TK
    lnTK = np.log(TK)
    Tsub = TK - 298.15

    # create blank parameter tables
    params = {k: np.zeros((N_CA, N_AN, *TK.shape)) for k in ['beta_0', 'beta_1', 'beta_2', 'C_phi']}

    # All except Table A8 - Temperature Sensitive
    for table in EQ_TABLES:
        # iterate through each parameter type and salt in each table.
        for (param, salt), g in TABLES[table].groupby(['Parameter', 'Salt']):
            p, n = break_salt(salt)  # identify which ions are involved
            if (p in CA_IND) and (n in AN_IND):
                # get the matrix indices of those ions
                pi = CA_IND[p]
                ni = AN_IND[n]
                
                eqn = EQ_TABLES[table]  # identify the correct equation
                if table in EQ_SPECIAL:  # does the table have special cases?
                    if salt in EQ_SPECIAL[table]:  # is this salt a special case?
                        eqn = EQ_SPECIAL[table][salt]  # if so, use the special equation.

                # calculate the parameter values and store them
                params[param][pi, ni] = eqn(a=g.values[0][2:], TK=TK, Tsub=Tsub, TKinv=TKinv, lnTK=lnTK)

    # Table A8 - Constants
    for i, row in TABLES['TabA8'].iterrows():
        p, n = break_salt(row.Salt)
        if (p in CA_IND) and (n in AN_IND):
            pi = CA_IND[p]
            ni = AN_IND[n]
            for param in params:
                params[param][pi, ni] = row[param]
    
    return params


def calc_Theta_Phi(TK):
    """
    Construct Theta and Phi matrices from Table A10 and A11 of Millero and Pierrot (1998).

    Parameters
    ----------
    TK : array-like
        Temperature in Kelvin
    
    Returns
    -------
    tuple of array-like
        Containing (Theta_negative, Theta_positive, Phi_NNP, Phi_PPN)
    """

    # create empty arrays
    Theta_positive = np.zeros((N_CA, N_CA, *TK.shape))
    Theta_negative = np.zeros((N_AN, N_AN, *TK.shape))
    Phi_PPN = np.zeros((N_CA, N_CA, N_AN, *TK.shape))
    Phi_NNP = np.zeros((N_AN, N_AN, N_CA, *TK.shape))

    # Assign static values from Table A11
    for _, row in TABA11.iterrows():
        ions = row.Pair.split('-')
        index = get_ion_index(row.Pair)

        if ions[0] in CA_IND:
            if len(ions) == 2:
                Theta_positive[index] = row.Value
                Theta_positive[index[::-1]] = row.Value
            elif len(ions) == 3:
                Phi_PPN[index] = row.Value
                Phi_PPN[index[1], index[0], index[2]] = row.Value
        
        if ions[0] in AN_IND:
            if len(ions) == 2:
                Theta_negative[index] = row.Value
                Theta_negative[index[::-1]] = row.Value
            elif len(ions) == 3:
                Phi_NNP[index] = row.Value
                Phi_NNP[index[1], index[0], index[2]] = row.Value

    # Assign T-sensitive values from Table A10
    pnames = ['a1', 'a2', 'a3_e4', 'a4_e4', 'a5_e6']  # parameter names in TABle
    for _, row in TABA10.iterrows():
        ions = row.Pair.split('-')
        index = get_ion_index(row.Pair)
        
        a = row[pnames]  # identify parameters
        val = EqA10(a, TK)  # calculate value
        
        # assign value
        if ions[0] in CA_IND:
            if len(ions) == 2:
                Theta_positive[index] = val
                Theta_positive[index[::-1]] = val
            elif len(ions) == 3:
                Phi_PPN[index] = val
                Phi_PPN[index[1], index[0], index[2]] = val
        if ions[0] in AN_IND:
            if len(ions) == 2:
                Theta_negative[index] = val
                Theta_negative[index[::-1]] = val
            elif len(ions) == 3:
                Phi_NNP[index] = val
                Phi_NNP[index[1], index[0], index[2]] = val

    # Special cases that deviate from values in Millero and Pierrot (1998)
    special = {
        'Na-Ca-Cl': -7.6398 + -1.2990e-2 * TK + 1.1060e-5 * TK**2 + 1.8475 * np.log(TK),  # Spencer et al 1990
        'Mg-Ca-Cl': 4.15790220e1 + 1.30377312e-2 * TK - 9.81658526e2 / TK - 7.4061986 * np.log(TK),  # Spencer et al 1990
        'Cl-CO3': -0.092,  #Spencer et al 1990
        'CO3-OH': 0.1,  # http://www.aim.env.uea.ac.uk/aim/accent4/parameters.html
    }

    for ionstr, v in special.items():
        ions = ionstr.split('-')
        index = get_ion_index(ionstr)
        if ions[0] in CA_IND:
            if len(ions) == 2:
                Theta_positive[index] = v
                Theta_positive[index[::-1]] = v
            elif len(ions) == 3:
                Phi_PPN[index] = v
                Phi_PPN[index[1], index[0], index[2]] = v
        if ions[0] in AN_IND:
            if len(ions) == 2:
                Theta_negative[index] = v
                Theta_negative[index[::-1]] = v
            elif len(ions) == 3:
                Phi_NNP[index] = v
                Phi_NNP[index[1], index[0], index[2]] = v

    return {
        'Theta_negative': Theta_negative, 
        'Theta_positive': Theta_positive, 
        'Phi_NNP': Phi_NNP, 
        'Phi_PPN': Phi_PPN
        }


def PitzerParams(TK):
    """
    Return Pitzer params for given T (Kelvin).
    
    Parameters
    ----------
    TK : array-like
        Temperature in Kelvin
        
    Returns
    -------
    dict of arrays
        with keys: beta_0, beta_1, beta_2, C_phi, Theta_negative, Theta_positive, Phi_NNP, Phi_PPN
    """
    if isinstance(TK, (float, int)):
        TK = np.asanyarray(TK)
    
    out = calc_beta_C(TK)
    out.update(calc_Theta_Phi(TK))

    return out



from curses.ascii import TAB
from distutils.command.build import build
from xml.etree.ElementTree import PI
from click import Tuple
import numpy as np
from .params import TABLES, Pind, Nind, Iind, N_anions, N_cations, get_ion_index  # constants
from .params import build_salt, break_salt
from .params import PitzerParams


def EqA1(a, TK, TKinv, lnTK, **kwargs):
    # T in Kelvin
    return (
        a[0] +
        a[1] * TK +
        a[2] * TKinv +
        a[3] * lnTK +
        a[4] / (TK - 263) +
        a[5] * TK**2 +
        a[6] / (680 - TK) +
        a[7] / (TK - 227)
    )

def EqA2(a, TK, **kwargs):
    return a[0] + a[1] * TK + a[2] * TK**2

def EqA3A4(a, Tsub, **kwargs):
    return a[0] + a[1] * Tsub + a[2] * Tsub**2

def EqA5A6(a, TK, Tsub, **kwargs):
    return a[0] + a[1] * Tsub + a[2] * (TK - 303.15)**2

def EqA7(a, TK, **kwargs):
    PR, PJ, PLR = a[0:3]
    return (
        PR + PJ * (8834524.63945833 - 88893.4225 * PLR) * (1 / TK - (1 / 298.15))
        + PJ / 6 * (TK**2 - 88893.4225)
    )

# link tables to equations
TabEqs = {
    'TabA1': EqA1,
    'TabA2': EqA2,
    'TabA3': EqA3A4,
    'TabA4': EqA3A4,
    'TabA5': EqA5A6,
    'TabA6': EqA5A6,
    'TabA7': EqA7,
    'TabA9': None,  # both special
    'TabSpecial': None  # special cases handled below
}

# Special Case Equations noted in table subscripts
def EqA2_MgSO4(a, TK, **kwargs):
    return (
        a[0] * ((TK / 2) + (88804) / (2 * TK) - 298)
        + a[1] * ((TK**2 / 6) + (26463592) / (3 * TK) - (88804 / 2))
        + a[2] * (TK**3 / 12 + 88804 * 88804 / (4 * TK) - 26463592 / 3)
        + a[3]
        * ((TK**4 / 20) + 88804 * 26463592 / (5 * TK) - 88804 * 88804 / 4)
        + a[4] * (298 - (88804 / TK))
        + a[5]
    )

def EqA3A4_XHS(a, Tsub, TKinv, **kwargs):
    return a[0] + a[1] * TKinv + a[2] * Tsub**2

def EqA3_XSO3(a, TK, TKinv):
    return a[0] + a[1] * (TKinv - 1/298.15) + a[2] * np.log(TK / 298.15)

def EqA9_HCl(a, TK, **kwargs):
    return a[0] + a[1] * TK + a[2] / TK

def EqA9_HSO4(a, TK, **kwargs):
    TKsub3 = TK - 328.15
    return (
        a[0] + TKsub3 * 1e-3 * 
        (a[1] + TKsub3 * ((a[2] / 2) + TKsub3 * (a[3] / 6)))
        )

# dictionary of special cases
EqSpecial = {
    'TabA2': {
        'MgSO4': EqA2_MgSO4
    },
    'TabA3': {
        'KHS': EqA3A4_XHS,
        'Na2SO3': EqA3_XSO3,
        'NaHSO3': EqA3_XSO3,
    },
    'TabA4': {
        'KHS': EqA3A4_XHS,
    },
    'TabA9': {
        'H-Cl': EqA9_HCl,
        'H-SO4': EqA9_HSO4
    },
}

# Special case equations used in Mathis Hain's code.
def EqSpecial_Na2SO4_Moller(a, TK, lnTK, **kwargs):
    return (
        a[0] +
        a[1] * TK +
        a[2] / TK +
        a[3] * lnTK +
        a[4] / (TK - 263) +
        a[5] * TK**2 +
        a[6] / (680. - TK)
    )

def EqSpecial_MgHSO4(a, Tsub, **kwargs):
    return a[0] + a[1] * Tsub

# add these to dictionary
EqSpecial['TabSpecial'] = {
        'Na2SO4': EqSpecial_Na2SO4_Moller,
        'Mg(HSO4)2': EqSpecial_MgHSO4
    }

def calc_beta_C(TK):
    """
    Calculate matrices of beta_0, beta_1, beta_2 and C_phi at given TK.

    Matrices are constructed from tables A1-A9 of Millero and Pierrot (1998; 
    doi:10.1023/A:1009656023546), with modifications implemented by Hain et 
    al (2015).

    Parameters
    ----------
    TK : array-like
        Temperature in Kelvin
    
    Returns
    -------
    dict
        Containing {beta_0, beta_1, beta_2, C_phi}
    """
    TKinv = 1. / TK
    lnTK = np.log(TK)
    Tsub = TK - 298.15

    # create blank parameter tables
    params = {k: np.zeros((N_cations, N_anions, *TK.shape)) for k in ['beta_0', 'beta_1', 'beta_2', 'C_phi']}

    # All except Table A8 - Temperature Sensitive
    for table in TabEqs:
        # iterate through each parameter type and salt in each table.
        for (param, salt), g in TABLES[table].groupby(['Parameter', 'Salt']):
            p, n = break_salt(salt)  # identify which ions are involved
            if (p in Pind) and (n in Nind):
                # get the matrix indices of those ions
                pi = Pind[p]
                ni = Nind[n]
                
                eqn = TabEqs[table]  # identify the correct equation
                if table in EqSpecial:  # does the table have special cases?
                    if salt in EqSpecial[table]:  # is this salt a special case?
                        eqn = EqSpecial[table][salt]  # if so, use the special equation.

                # calculate the parameter values and store them
                params[param][pi, ni] = eqn(a=g.values[0][2:], TK=TK, Tsub=Tsub, TKinv=TKinv, lnTK=lnTK)

    # Table A8 - Constants
    for i, row in TABLES['TabA8'].iterrows():
        p, n = break_salt(row.Salt)
        if (p in Pind) and (n in Nind):
            pi = Pind[p]
            ni = Nind[n]
            for param in params:
                params[param][pi, ni] = row[param]
    
    return params


# # testing
# TK = np.array([15, 25, 35]) + 298.15

# params = calc_beta_C(TK)

# test_params = PitzerParams(T=TK)

# Prind = {v:k for k, v in Pind.items()}
# Nrind = {v:k for k, v in Nind.items()}

# def compare_table(new, ref):
#     ind = np.abs(new - ref) > 1e-12
#     locs = np.argwhere(ind)
    
#     for loc in locs:
#         loc = tuple(loc)
#         iP = Prind[loc[0]]
#         iN = Nrind[loc[1]]
#         iTK = TK[loc[2]]

#         print(f'  {iP}-{iN} at {iTK}: {ref[loc]} vs {new[loc]}')

#     # print(ref[Tuple(loc)])

# for p in ['beta_0', 'beta_1', 'beta_2', 'C_phi']:
#     diff = np.abs(params[p] - test_params[p])
#     if np.all(diff < 1e-10):
#         print(f'{p}:   OK')
#     else:
#         print(p)
#         compare_table(params[p], test_params[p])


# if __name__ == "__main__":
    # print('yup')


import os
import unittest
import numpy as np
import pandas as pd

import pymyami

class MyAMI_V1_crosscheck(unittest.TestCase):

    def test_Fcorr(self):
        checkfile = os.path.join(os.path.dirname(__file__), 'data/MyAMI_V1_Fcorr_checkvals.csv')
        print(checkfile)
        check = pd.read_csv(checkfile, index_col=0)
        check.columns = ['T', 'S', 'Ca', 'Mg', 'KspC', 'K1', 'K2', 'KW', 'KB', 'KspA', 'K0', 'KS']

        new_Fcorr = pymyami.calc_Fcorr(Sal=check.S.values, TempC=check['T'].values, Ca=check.Ca.values, Mg=check.Mg.values)

        Ks = 'K0', 'K1', 'K2', 'KW', 'KB', 'KspA', 'KspC', 'KS'

        print('Comparing Fcorr to MyAMI_V1 (must be <0.4% max difference)')
        for k in Ks:
            v1 = check[k]
            new = new_Fcorr[k]

            rdiff = (v1 - new) / v1  # relative difference
            
            maxpercentdiff = 100 * np.max(np.abs(rdiff))
            avgpercentdiff = 100 * np.mean(rdiff)
            
            print(f'  {k}: {maxpercentdiff:.2f}% max, {avgpercentdiff:.2f}% avg')
            self.assertLess(maxpercentdiff, 0.4, msg=f'Maximum difference in {k} correction factor too large: {maxpercentdiff}%')

if __name__ == "__main__":
    unittest.main()



""""
User-facing function for calculating K correction factors using MyAMI
"""

from .helpers import shape_matcher
from .pitzer import calculate_gKs
from .params import PitzerParams


def calc_Fcorr(Sal=35., TempC=25., Na=None, K=None, Mg=None, Ca=None, Sr=None, Cl=None, BOH4=None, HCO3=None, CO3=None, SO4=None):
    """
    Calculate K correction factors as a fn of temp and salinity that can be applied to empirical Ks

    Parameters
    ----------
    TempC : array-like
        Temperature in Celcius
    Sal : array-like
        Salinity in PSU
    Na, K, Mg, Ca, Sr : array-like
        Cation concentrations in mol/kg. If None, mean ocean water values are used.
    Cl, BOH4, HCO3, CO3, SO4 : array-like
        Anion concentrations in mol/kg. If None, mean ocean water values are used.


    Returns
    -------
    dict 
        Correction factors (Fcorr) to be applied to empirical K values,
        where K_corr = K_cond * F_corr.
    """

    # ensure all inputs are the same shape
    TempC, Sal, Na, K, Mg, Ca, Sr, Cl, BOH4, HCO3, CO3, SO4 = shape_matcher(TempC, Sal, Na, K, Mg, Ca, Sr, Cl, BOH4, HCO3, CO3, SO4)

    # Calculate pitzer parameters at given temperature
    pitzer_params = PitzerParams(TempC + 273.15)

    # Calculate gK's for modern (mod) and experimental (x) seawater composition
    (
        gKspC_mod,
        gK1_mod,
        gK2_mod,
        gKW_mod,
        gKB_mod,
        gKspA_mod,
        gK0_mod,
        gKS_mod,
    ) = calculate_gKs(TempC, Sal, **pitzer_params)
    
    (
        gKspC_X, 
        gK1_X, 
        gK2_X, 
        gKW_X, 
        gKB_X, 
        gKspA_X, 
        gK0_X, 
        gKS_X) = calculate_gKs(
            TempC, Sal, Na=Na, K=K, Mg=Mg, Ca=Ca, Sr=Sr, Cl=Cl, BOH4=BOH4, HCO3=HCO3, CO3=CO3, SO4=SO4, **pitzer_params)

    # Calculate conditional K's predicted for seawater composition X
    F_dict = {
        "K0": gK0_X / gK0_mod,
        "K1": gK1_X / gK1_mod,
        "K2": gK2_X / gK2_mod,
        "KB": gKB_X / gKB_mod,
        "KW": gKW_X / gKW_mod,
        "KspC": gKspC_X / gKspC_mod,
        "KspA": gKspA_X / gKspA_mod,
        "KS": gKS_X / gKS_mod,
    }
    
    return F_dict

